aliases = ['l']


def generate(args):
    if len(args) != 1:
        msg = 'The field type "list" takes a single argument. i.e., a comma-separated list'
        raise ValueError(msg)
    # ---
    return list(filter(len, args[0].split(',')))



import sys
import time
import types
import signal
import logging
import subprocess

from enum import Enum
from threading import Timer

from . import brlogger, brconsole
from .lib import Config, CLISyntaxError
from .utils import Pool
from .constants import *
from .console import restrict_console_access
from .exceptions import TaskFailureError


class AppStatus(Enum):
    INITIALIZING = 1
    RUNNING = 2
    ABORTING = 3
    KILLING = 4
    DONE = 10


class Brun():

    _status = AppStatus.INITIALIZING
    _sigint_counter = 0
    _sigint_received_time = None

    def __init__(self, args):
        self.status(AppStatus.INITIALIZING)
        self._setup_signal_handler()
        self.args = args
        # ---
        # configure logger
        if self.args.suppress_warnings:
            brlogger.addFilter(lambda e: e.levelno != logging.WARNING)
        if self.args.debug:
            brlogger.setLevel(logging.DEBUG)
        if self.args.suppress_warnings and self.args.debug:
            brlogger.info('Warnings cannot be suppressed when --debug is active.')
        # configure console
        if not self.args.debug:
            restrict_console_access(brlogger)
        brconsole.set_show_status(not self.args.no_status)
        brconsole.set_plain_console(self.args.debug)
        # turn fields and groups into lists
        if 'field' in self.args:
            self.args.field = [self.args.field] \
                if not isinstance(self.args.field, list) else self.args.field
        self.args.group = [self.args.group] \
            if not isinstance(self.args.group, list) else self.args.group
        # parse brun configuration
        try:
            self.config = Config(self.args)
        except CLISyntaxError as e:
            brlogger.error(str(e))
            exit(-1)
        except Exception as e:
            brlogger.error(str(e))
            exit(-2)
        # define number of workers
        num_workers = self.args.parallel if self.args.parallel != -1 else len(self.config)
        num_workers = min(MAX_PARALLEL_WORKERS, max(1, num_workers))
        self.is_parallel = num_workers > 1
        # create workers pool
        self.pool = Pool(num_workers, self._exception_handler)

    def start(self):
        self.status(AppStatus.RUNNING)
        # add commands to the pool
        for cc in self.config:
            cmd = cc.apply(self.args.command)
            self.pool.enqueue(self._worker_task, cmd)
        # start pool
        self.pool.run()
        # monitor the status of the app
        while (self.pool.alive() and not self.pool.idle()) or (not self.pool.done()):
            self._update_status()
            brconsole.set_progress(self._get_progress())
            # Status: ABORTING
            if self.status() == AppStatus.ABORTING:
                self.pool.abort()
            # Status: KILLING
            if self.status() == AppStatus.KILLING:
                brlogger.warning('Escalating to KILL...')
                sys.exit(1)
            # breath
            time.sleep(1.0 / APP_HEARTBEAT_HZ)
        # update status bar one more time and then stop it
        brconsole.set_progress(self._get_progress())
        brconsole.set_show_status(False)
        # show collected errors
        self._process_tasks_output(stderr_only=(not self.args.verbose))
        # ---
        brlogger.info('Done!')
        brconsole.close()

    def status(self, status=None):
        if status and not isinstance(status, AppStatus):
            raise ValueError('Invalid status {}'.format(status))
        if status:
            self._status = status
        return self._status

    def abort(self):
        self.status(AppStatus.ABORTING)

    def kill(self):
        self.status(AppStatus.KILLING)

    def _update_status(self):
        # escalate SIGINT -> SIGKILL ater some time
        s = self._sigint_received_time
        if s and time.time() - s > ESCALATE_TO_KILL_AFTER_SECS:
            self.status(AppStatus.KILLING)

    def _process_tasks_output(self, stderr_only=False):
        cmds = []
        stdouts = []
        stderrs = []
        # get cmd, stdout, and stderr
        for res in self.pool.results():
            cmd = ' '.join(res.cmd)
            cmds.append(cmd)
            stdouts.append(res.stdout)
            stderrs.append(res.stderr)
        # dump stdout to console
        if not stderr_only:
            for cmd, std in zip(cmds, stdouts):
                if res.returncode == 0 and len(std):
                    brconsole.write(TASK_OUTPUT_TEMPLATE.format(cmd=cmd, content=std.rstrip()))
        # dump stderr to console
        for cmd, std in zip(cmds, stderrs):
            if res.returncode != 0 and len(std):
                brconsole.write(TASK_ERROR_TEMPLATE.format(cmd=cmd, content=std.rstrip()))

    def _get_progress(self):
        stats = self.pool.get_stats()
        stats['app_status'] = {
            AppStatus.INITIALIZING: 'init',
            AppStatus.RUNNING: 'healthy',
            AppStatus.ABORTING: 'aborting',
            AppStatus.KILLING: 'killing'
        }[self.status()]
        return stats

    def _worker_task(self, cmd):
        cmd_str = ' '.join(cmd)
        stdout = subprocess.PIPE if self.is_parallel else sys.stdout
        result = types.SimpleNamespace(cmd=cmd, stdout="", stderr="", returncode=None)
        # -->
        brlogger.info(PARALLEL_TO_START_PROMPT_STRING[self.is_parallel].format(cmd_str))
        if not self.args.dry_run:
            no_sigint = lambda: signal.signal(signal.SIGINT, signal.SIG_IGN)
            # launch task
            task = subprocess.Popen(cmd_str,
                                    shell=True,
                                    stdout=stdout,
                                    stderr=subprocess.PIPE,
                                    preexec_fn=no_sigint)
            # wait for the task to end
            task.wait()
            if self.is_parallel:
                result.stdout = task.stdout.read().decode('utf-8')
            result.stderr = task.stderr.read().decode('utf-8')
            result.returncode = task.returncode
            # on failure
            if task.returncode != 0:
                brlogger.info(PARALLEL_TO_FAILURE_PROMPT_STRING[self.is_parallel].format(cmd_str))
                # raise error
                msg = 'The command {} failed with exit code {}.'
                raise TaskFailureError(msg.format(cmd, task.returncode), result)
        # <--
        brlogger.info(PARALLEL_TO_END_PROMPT_STRING[self.is_parallel].format(cmd_str))
        return result

    def _exception_handler(self, name, exception_type, exception, tback, *args, **kwargs):
        # abort remaining tasks
        if not self.args.ignore_errors:
            self.abort()

    def _setup_signal_handler(self):
        # create signal handler function
        def signal_handler(sig, frame):
            self._sigint_counter += 1
            if self.status() == AppStatus.INITIALIZING:
                # just exit if still initializing
                sys.exit(0)
            if self.status() == AppStatus.RUNNING:
                # shutdown app (if running)
                brlogger.warning(
                    ' Request of interruption received. Waiting for tasks to finish...')
                brlogger.warning(' (Press Ctrl+C three times to force kill)')
                self.abort()
            elif self.status() == AppStatus.ABORTING and self._sigint_counter == 3:
                # kill app if already aborting
                self.kill()

        # register signal handler
        signal.signal(signal.SIGINT, signal_handler)



aliases = ['r']


def generate(args):
    if len(args) != 1:
        msg = 'The field type "range" takes a single argument. i.e., a range descriptor'
        raise ValueError(msg)
    # ---
    vals_str = args[0].split(',')
    vals = [int(v) for v in vals_str]
    fmt = _extract_format(vals_str)
    if len(vals) == 1:
        return _format(range(vals[0]), fmt)
    elif len(vals) == 2:
        return _format(range(vals[0], vals[1], 1), fmt)
    elif len(vals) == 3:
        return _format(range(vals[0], vals[1], vals[2]), fmt)
    else:
        msg = 'The argument for field type "range" can be of the forms:' + \
                '   - f  ->  [0,1,...,f)' + \
                '   - s,f  ->  [s,s+1,...,f)' + \
                '   - s,f,i  ->  [s,s+i,...,f)'
        raise ValueError(msg)


def _extract_format(vals_str):
    args = vals_str + [vals_str[0]]
    s, f = (args[0], args[1]) if int(args[1]) > int(args[0]) else (args[1], args[0])
    if len(s) != len(str(int(s))):
        padding_len = len(s) - len(str(int(s)))
        padding = s[:padding_len]
        if len(set(padding)) != 1:
            msg = 'Formatted arguments in field "range" must have a single char padding'
            raise ValueError(msg)
        padding_char = padding[0]
        return "{:%s%sd}" % (padding_char, len(s))
    return "{0}"


def _format(lst, fmt):
    return list(map(lambda v: fmt.format(v), lst))



import os
import io
import sys
import time
import logging
import threading

from collections import defaultdict

from queue import Queue


def restrict_console_access(logger):
    sys.stdout = ProxyStream(logger, logging.INFO)
    sys.stderr = ProxyStream(logger, logging.ERROR)


class ProxyStream(io.TextIOWrapper):
    def __init__(self, logger, level, *args, **kwargs):
        self._logger = logger
        self._level = level
        self._buffer = ""
        self._lock = threading.Semaphore(1)
        self._disabled = False
        super(ProxyStream, self).__init__(open(os.devnull, 'w'), *args, **kwargs)

    def write(self, s, *args, **kwargs):
        if self._disabled:
            return
        self._lock.acquire()
        self._buffer += s
        lines = self._buffer.split('\n')
        for line in lines[:-1]:
            self._logger._log(self._level, line, None)
        self._buffer = lines[-1]
        self._lock.release()


class Console():
    def __init__(self):
        # redirect stdout to buffer
        self.start_time = time.time()
        self.stdout = sys.stdout
        self.buffer = Queue()
        self.lock = threading.Semaphore(1)
        self.progress = defaultdict(lambda: 0)
        self.has_progress = False
        self.delete_last = False
        self.show_status = None
        self.plain_console = False

    def set_progress(self, progress):
        self.progress.update(progress)
        self.has_progress = True
        self.flush()

    def uptime(self):
        return time.time() - self.start_time

    def set_show_status(self, val):
        self.show_status = val

    def set_plain_console(self, val):
        self.plain_console = val

    def write(self, msg):
        # get the lock
        self.lock.acquire()
        # write to buffer
        self.buffer.put(msg)
        # release the lock
        self.lock.release()

    def flush(self):
        # get the lock
        self.lock.acquire()
        # clear last line
        self._clear_status_bar()
        # dump buffer content on the console
        while not self.buffer.empty():
            line = self.buffer.get()
            self.stdout.write(line + "\033[K")
        # print separator then status bar
        self._print_status_bar()
        # flush stdout
        self.stdout.flush()
        # release the lock
        self.lock.release()

    def close(self):
        self.set_show_status(False)
        # clear status (if any)
        self._clear_status_bar()

    def _clear_status_bar(self):
        if self.delete_last and (not self.plain_console):
            # clear 2 lines
            self.stdout.write("\033[F\033[K" * 2)
            self.delete_last = False
            # flush stdout
            self.stdout.flush()

    def _print_status_bar(self):
        if self.show_status and self.has_progress:
            self.stdout.write(("-" * 80) + "\n")
            self.stdout.write(self._get_status_bar() + "\n")
            self.delete_last = True
            # flush stdout
            self.stdout.flush()

    def _get_status_bar(self):
        return "[brun {:.1f} s] [{:s}] [{:d}/{:d} complete] [{:d}/{:d} jobs] [{:d} queued] [{:d} aborted] [{:d} failed]".format(
            self.uptime(),
            str(self.progress['app_status']),
            self.progress['tasks_completed'],
            self.progress['tasks_total'],
            self.progress['jobs_max'] - self.progress['jobs_idle'],
            self.progress['jobs_max'],
            self.progress['tasks_queued'],
            self.progress['tasks_aborted'],
            self.progress['tasks_failed'],
        )



# generator: list
from .list import \
    generate as generator_list, \
    aliases as aliases_list

# generator: range
from .range import \
    generate as generator_range, \
    aliases as aliases_range

# generator: file
from .file import \
    generate as generator_file, \
    aliases as aliases_file

# generator: stdin
from .stdin import \
    generate as generator_stdin, \
    aliases as aliases_stdin

# generator: glob
from .glob import \
    generate as generator_glob, \
    aliases as aliases_glob

# generator: json
from .json import \
    generate as generator_json, \
    aliases as aliases_json

_generators_map = {
    'list': generator_list,
    'range': generator_range,
    'file': generator_file,
    'stdin': generator_stdin,
    'glob': generator_glob,
    'json': generator_json,
}

_generators_map.update({alias: generator_list for alias in aliases_list})
_generators_map.update({alias: generator_range for alias in aliases_range})
_generators_map.update({alias: generator_file for alias in aliases_file})
_generators_map.update({alias: generator_stdin for alias in aliases_stdin})
_generators_map.update({alias: generator_glob for alias in aliases_glob})
_generators_map.update({alias: generator_json for alias in aliases_json})


def _get_generator(gen_name):
    if not gen_name in _generators_map:
        msg = 'Field type "{}" not found!'.format(gen_name)
        raise ValueError(msg)
    # ---
    return _generators_map[gen_name]



from queue import Queue
from threading import Thread, Event, Semaphore
from sys import stdout, stderr, exc_info
from time import sleep
from collections import defaultdict
from copy import copy

import traceback

from brun import brlogger
from brun.exceptions import TaskFailureError


class Worker(Thread):
    """Thread executing tasks from a given tasks queue"""
    def __init__(self, name, queue, results, abort, idle, exception_handler, stats):
        Thread.__init__(self)
        self.name = name
        self.queue = queue
        self.results = results
        self.abort = abort
        self.idle = idle
        self.exception_handler = exception_handler
        self.stats = stats
        self.daemon = True
        self.start()

    """Thread work loop calling the function with the params"""

    def run(self):
        #keep running until told to abort
        while not self.abort.is_set():
            try:
                #get a task and raise immediately if none available
                func, args, kwargs = self.queue.get(False)
                self.idle.clear()
            except:
                #no work to do
                self.idle.set()
                sleep(0.5)
                continue

            try:
                #the function may raise
                result = func(*args, **kwargs)
                self.stats.increase('tasks_completed')
                if (result is not None):
                    self.results.put(result)
            except TaskFailureError:
                # get info about the error
                ex_type, ex, tb = exc_info()
                # get partial results and errors
                result = ex.result
                if (result is not None):
                    self.results.put(result)
                #so we move on and handle it in whatever way the caller wanted
                self.stats.increase('tasks_failed')
                self.exception_handler(self.name, ex_type, ex, tb, args, kwargs)
            except:
                ex_type, ex, tb = exc_info()
                self.stats.increase('tasks_failed')
                traceback.print_exception(ex_type, ex, tb, file=sys.stderr)
            finally:
                #task complete no matter what happened
                self.queue.task_done()


class Pool:
    """Pool of threads consuming tasks from a queue"""
    def __init__(self, thread_count, exception_handler):
        #batch mode means block when adding tasks if no threads available to process
        self.queue = Queue()
        self.resultQueue = Queue()
        self.thread_count = thread_count
        self.exception_handler = exception_handler
        self.stats = StatisticsCollector()
        self.aborts = []
        self.idles = []
        self.threads = []

    """Tell my threads to quit"""

    def __del__(self):
        self.abort()

    """Start the threads, or restart them if you've aborted"""

    def run(self, block=False):
        #either wait for them to finish or return false if some arent
        if block:
            while self.alive():
                sleep(1)
        elif self.alive():
            return False

        #go start them
        self.aborts = []
        self.idles = []
        self.threads = []
        for n in range(self.thread_count):
            abort = Event()
            idle = Event()
            self.aborts.append(abort)
            self.idles.append(idle)
            self.threads.append(
                Worker('thread-%d' % n, self.queue, self.resultQueue, abort, idle,
                       self.exception_handler, self.stats))
        return True

    """Add a task to the queue"""

    def enqueue(self, func, *args, **kargs):
        self.queue.put((func, args, kargs))
        self.stats.increase('tasks_total')

    """Wait for completion of all the tasks in the queue"""

    def join(self):
        self.queue.join()

    """Tell each worker that its done working"""

    def abort(self, block=False):
        #tell the threads to stop after they are done with what they are currently doing
        for a in self.aborts:
            a.set()
        # clear the queue
        while not self.queue.empty():
            try:
                _, args, _ = self.queue.get(False)
                cmd, *_ = args
                self.queue.task_done()
                brlogger.info(':brun: Aborted < {0}'.format(' '.join(cmd)))
                self.stats.increase('tasks_aborted')
            except:
                pass
        #wait for them to finish if requested
        while block and self.alive():
            sleep(1)

    """Returns True if any threads are currently running"""

    def alive(self):
        return True in [t.is_alive() for t in self.threads]

    """Returns True if all threads are waiting for work"""

    def idle(self):
        return False not in [i.is_set() for i in self.idles]

    """Returns True if not tasks are left to be completed"""

    def done(self):
        return self.queue.empty()

    """Get the set of results that have been processed, repeatedly call until done"""

    def results(self, wait=0):
        sleep(wait)
        results = []
        try:
            while True:
                #get a result, raises empty exception immediately if none available
                results.append(self.resultQueue.get(False))
                self.resultQueue.task_done()
        except:
            pass
        return results

    """Wait for the pool to complete and return the results as soon as they are ready"""

    def iterate_results(self):
        while not self.done() or not self.idle():
            for r in self.results():
                yield r
        for r in self.results():
            yield r

    def get_stats(self):
        stats = self.stats.get_stats()
        stats['jobs_idle'] = len([1 for i in self.idles if i.is_set()])
        stats['jobs_max'] = len([1 for t in self.threads if t.is_alive()])
        stats['tasks_queued'] = self.queue.qsize()
        return stats


class StatisticsCollector():
    def __init__(self):
        self.lock = Semaphore(1)
        self.data = defaultdict(lambda: 0)

    def set(self, key, value):
        self.lock.acquire()
        self.data[key] = value
        self.lock.release()

    def increase(self, key):
        self.lock.acquire()
        self.data[key] += 1
        self.lock.release()

    def decrease(self, key):
        self.lock.acquire()
        self.data[key] -= 1
        self.lock.release()

    def get_stats(self):
        self.lock.acquire()
        stats = copy(self.data)
        self.lock.release()
        return stats



import os

aliases = ['f']


def generate(args):
    if len(args) != 1:
        msg = 'The field type "file" takes a single argument. i.e., the path to a file'
        raise ValueError(msg)
    # ---
    filepath = args[0]
    if not os.path.exists(filepath):
        msg = 'The file "{}" does not exist.'.format(filepath)
        raise ValueError(msg)
    # ---
    with open(filepath, 'rt') as fin:
        lines = fin.readlines()
        return [line.strip() for line in lines]



import sys

aliases = []


def generate(args):
    if len(args) > 1:
        msg = 'The field type "stdin" takes one optional argument. i.e., a string delimiter (default=\\n)'
        raise ValueError(msg)
    # ---
    sep = '\n'
    if len(args) > 0:
        sep = args[0].strip()
    values = [l.strip().replace('\n', ';') for l in sys.stdin.read().split(sep)]
    # ---
    return list(filter(len, values))



class CLISyntaxError(RuntimeError):
    def __init__(self, msg):
        super(RuntimeError, self).__init__(msg)


class TaskFailureError(RuntimeError):
    def __init__(self, msg, result=None):
        self.result = result
        super(RuntimeError, self).__init__(msg)


class InvalidConfigurationError(RuntimeError):
    def __init__(self, msg, data=None):
        self.data = data
        super(RuntimeError, self).__init__(msg)



from brun import brlogger


def combine(lst1, lst2, _):
    if len(lst1) != len(lst2):
        brlogger.warning('Zipping two sets of different sizes {} != {}'.format(
            len(lst1), len(lst2)))
    return zip(lst1, lst2)



import os
import re
import itertools
import networkx as nx

from . import brlogger
from .exceptions import CLISyntaxError, InvalidConfigurationError
from .generators import _get_generator
from .combinators import _get_combinator
from .constants import *


class CommandConfig(object):
    def __init__(self, dp_dict=None):
        self._data = dp_dict if dp_dict else dict()

    def set(self, key, value):
        self._data[key] = value

    def get(self, key):
        return self._data[key]

    def apply(self, command):
        try:
            return [c.format(**self._data) for c in command]
        except KeyError as e:
            msg = 'The command contains a field {} that was not declared'.format(e)
        raise CLISyntaxError(msg)

    def __str__(self):
        return self._data.__str__()


class Config(object):
    def __init__(self, parsed):
        self._data = []
        self._fields = dict()
        # use default field if none were given
        if 'field' not in parsed:
            default_json_input_file = os.path.join(os.getcwd(), DEFAULT_FIELD[2])
            if os.path.exists(default_json_input_file):
                parsed.field = [':'.join(DEFAULT_FIELD)]
            else:
                raise CLISyntaxError(
                    "Argument --field/-f is required when the file '.brun' does not exist.")
        # parse fields and create their data points
        for field_str in parsed.field:
            # parse field
            name, type, generator_args = _parse_field(field_str)
            # make sure the fields are unique
            if name in self._fields:
                msg = 'Collision between fields. ' + \
                      'The field {} was declared more than once.'.format(name)
                raise ValueError(msg)
            # generate data
            generator = _get_generator(type)
            values = generator(generator_args)
            if isinstance(values, dict):
                self._fields.update(values)
                # fields that are generated together are naturally grouped using zip
                keys = list(values.keys())
                for k0, k1 in zip(keys, keys[1:]):
                    parsed.group.append('zip:{},{}'.format(k0, k1))
            else:
                self._fields[name] = values
        self._fields_keys = sorted(list(self._fields.keys()))
        # create fields graph
        G = nx.Graph()
        # add nodes
        G.add_nodes_from(self._fields_keys)
        brlogger.debug('Add nodes {}'.format(self._fields_keys))
        # parse groups
        for group_str in parsed.group:
            # parse group
            type, fields, combinator_args = _parse_group(group_str, self._fields_keys)
            # add edges
            for f0, f1 in zip(fields, fields[1:]):
                brlogger.debug('Add edge. Type {} between fields ({}, {})'.format(type, f0, f1))
                G.add_edge(f0, f1, type=type, args=combinator_args)
        # make sure the graph is consistent
        head, tail = _complete_graph(G)
        if not head:
            return
        # get path from head to tail
        paths = list(nx.all_simple_paths(G, source=head, target=tail)) or [[head]]
        if len(paths) > 1:
            raise InvalidConfigurationError('The induced graph has more than one path')
        self._fields_keys = paths[0]
        # combine fields
        data = [(v, ) for v in self._fields[self._fields_keys[0]]] if self._fields_keys else []
        field_to_blob = {f: self._fields[f] for f in self._fields_keys}
        for u, v in zip(self._fields_keys, self._fields_keys[1:]):
            combinator_data = G.get_edge_data(u, v)
            combinator_args = combinator_data['args']
            combinator = _get_combinator(combinator_data['type'])
            blob0, blob1 = field_to_blob[u], field_to_blob[v]
            data = field_to_blob[v] = _flatten_data(combinator(blob0, blob1, combinator_args))
        # turn data into CommandConfigs
        for d in data:
            assert len(self._fields_keys) == len(d)
            cc_dict = {k: v for k, v in zip(self._fields_keys, d)}
            cc = CommandConfig(cc_dict)
            self._data.append(cc)

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)


def _parse_field(field_str):
    parts = tuple(field_str.split(':'))
    if len(parts) <= 1:
        raise CLISyntaxError('Syntax error in field descriptor "{}"'.format(field_str))
    # ---
    name, type, *generator_args = parts
    # validate field name
    if not re.search('\w+', name):
        raise CLISyntaxError(
            'Field name {} not valid. Only letters and numbers are allowed'.format(name))
    # ---
    return name, type, generator_args


def _parse_group(group_str, declared_fields):
    parts = tuple(group_str.split(':'))
    if len(parts) <= 1:
        raise CLISyntaxError('Syntax error in group descriptor "{}"'.format(group_str))
    # ---
    type, fields, *combinator_args = parts
    # verify fields
    fields = fields.split(',')
    if len(fields) <= 1:
        msg = 'Group descriptor "{}" should have at least two fields, {} given'.format(
            group_str, len(fields))
        raise CLISyntaxError(msg)
    for field in fields:
        if field not in declared_fields + ['*']:
            msg = 'Field "{}" used in group argument "{}" was not declared'.format(
                field, group_str)
            raise CLISyntaxError(msg)
    # ---
    return type, fields, combinator_args


def _flatten_data(blob):
    nblob = []
    for e in blob:
        if isinstance(e, (tuple, list)) and isinstance(e[0], (tuple, list)):
            nblob.append(e[0] + (e[1], ))
        else:
            nblob.append(e)
    return nblob


def _complete_graph(G, default_comb=DEFAULT_COMBINATOR):
    """
    The given groups have added edges to the graph to form several connected components.
    A graph is valid iff all connected components are valid.
    A connected component is valid iff:
        - it has n vertices and exactly n-1 edges
        - the degree of each node is either 1 or 2
    """
    cc = [G.subgraph(c).copy() for c in nx.connected_components(G)]
    cc_degree = [G1.degree() for G1 in cc]
    unitary_nodes = []
    for G1, degree in zip(cc, cc_degree):
        msg = 'Invalid grouping. No loops of forks are allowed. ' + \
              'Invalid grouping found between the fields: {}'.format(G1.nodes)
        e = InvalidConfigurationError(msg, data=set(G1.nodes))
        # check if this connected component is valid
        if G1.number_of_edges() != G1.number_of_nodes() - 1:
            raise e
        for _, d in degree:
            if d not in [0, 1, 2]:
                raise e
        ends = [n for n, d in degree if d in [0, 1]]
        unitary_nodes += ends if len(ends) > 1 else 2 * [ends[0]]
    # all the connected components are valid, add missing edges
    for n0, n1 in zip(unitary_nodes[1:-2], unitary_nodes[2:-1]):
        G.add_edge(n0, n1, type=default_comb, args=None)
    # return head, tail
    return (unitary_nodes[0], unitary_nodes[-1]) if len(unitary_nodes) else (None, None)



import os
import glob

aliases = ['g']

type_to_fcn_map = {
    '*': lambda e: True,
    'f': os.path.isfile,
    'd': os.path.isdir,
}


def generate(args):
    args = args[0].split(',')
    if len(args) not in [1, 2, 3]:
        raise ValueError('The field type "glob" takes the following arguments: ' + \
                         'path[,query[,filter-type]]'
        )
    # ---
    path = args[0]
    if not os.path.exists(path):
        raise ValueError('The path "{}" does not exist.'.format(path))
    if not os.path.isdir(path):
        raise ValueError('The path "{}" is not a directory.'.format(path))
    # ---
    query = '*' if len(args) <= 1 else args[1]
    # ---
    type = '*' if len(args) <= 2 else args[2].strip()
    if type not in type_to_fcn_map:
        raise ValueError('Unknown type "{}" for filter-type argument. Allowed {}'.format(
            type, list(type_to_fcn_map.keys())))
    filter_type = type_to_fcn_map[type]
    # ---
    glob_query = os.path.join(path, query)
    return sorted([os.path.basename(p) for p in glob.glob(glob_query) if filter_type(p)])



import itertools


def combine(lst1, lst2, args):
    if len(args) != 1:
        msg = 'The grouping strategy "zipl" requires a fill argument'
        raise ValueError(msg)
    fill = args[0]
    return itertools.zip_longest(lst1, lst2, fillvalue=fill)



import sys
import argparse

from .main import Brun
from . import brlogger, __version__
from .constants import DEFAULT_COMBINATOR


def run():
    parser = get_parser()
    # ---
    # version
    f = sys.argv.index('--') if '--' in sys.argv else len(sys.argv)
    if '-v' in sys.argv[1:f] or '--version' in sys.argv[1:f]:
        print('brun version {}\n'.format(__version__))
        exit(0)
    # ---
    # parse arguments
    parsed = parser.parse_args()
    # create app and spin it
    app = Brun(parsed)
    app.start()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f',
                        '--field',
                        action='append',
                        default=argparse.SUPPRESS,
                        help="Specify a field (syntax: 'name:type[:args]')")
    parser.add_argument('-g',
                        '--group',
                        action='append',
                        default=[],
                        help="Group two or more fields together according to a " + \
                        "combination strategy (syntax: 'strategy:field1,field2[,...]'). " + \
                        "Default is {}".format(DEFAULT_COMBINATOR)
    )
    parser.add_argument('-p',
                        '--parallel',
                        const=-1,
                        default=1,
                        action='store',
                        nargs='?',
                        type=int,
                        help="How many commands can run in parallel")
    parser.add_argument('-i',
                        '--interactive',
                        action='store_true',
                        default=False,
                        help="Whether to run the commands in interactive mode")
    parser.add_argument('-D',
                        '--dry-run',
                        action='store_true',
                        default=False,
                        help="Performs a dry-run. It shows which commands would run")
    parser.add_argument('--debug', action='store_true', default=False, help="Run in debug mode")
    parser.add_argument('-vv',
                        '--verbose',
                        dest='verbose',
                        action='store_true',
                        default=False,
                        help="Run in verbose mode")
    parser.add_argument('--suppress-warnings',
                        action='store_true',
                        default=False,
                        help="Suppress warnings")
    parser.add_argument('-I',
                        '--ignore-errors',
                        action='store_true',
                        default=False,
                        help="Do not stop if one command fails")
    parser.add_argument('--no-status',
                        action='store_true',
                        default=False,
                        help="Do not show the status bar")
    parser.add_argument('command', nargs='+')
    return parser



import os
import sys
import time
import logging
import threading

from queue import Queue


def get_logger(console):
    # create logger
    logging.basicConfig()
    logger = BrLogger(console, 'brun', logging.INFO)
    return logger


class BrLogger(logging.Logger):
    def __init__(self, console, name, level=logging.NOTSET):
        super(BrLogger, self).__init__(name, level)
        self.console = console

    def _log(self, lvl, msg, *args, **kwargs):
        # build message
        if not msg.endswith('\n'):
            msg += '\n'
        msg = '{}:{}'.format(logging._levelToName[lvl], msg)
        # add message to buffer
        self.console.write(msg)
        # force flush
        self.console.flush()



import sys
from .logger import get_logger
from .console import Console

__version__ = '0.2.3'

# create console
brconsole = Console()

# get logger
brlogger = get_logger(brconsole)

# import CLI entrypoint
from .cli import run



from .cross import combine as combine_cross
from .zip import combine as combine_zip
from .zipl import combine as combine_zipl

_combinator_map = {
    'default': combine_cross,
    'cross': combine_cross,
    'zip': combine_zip,
    'zipl': combine_zipl,
}


def _get_combinator(comb_name):
    if not comb_name in _combinator_map:
        msg = 'Grouping strategy "{}" not found!'.format(comb_name)
        raise ValueError(msg)
    # ---
    return _combinator_map[comb_name]



import sys
from distutils.core import setup


def get_version(filename):
    import ast
    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError('No version found in %r.' % filename)
    if version is None:
        raise ValueError(filename)
    return version


if sys.version_info < (3, 5):
    msg = 'run-batch works with Python 3.5 and later.\nDetected %s.' % str(sys.version)
    sys.exit(msg)

lib_version = get_version(filename='brun/__init__.py')

setup(
    name='brun',
    packages=[
        'brun',
        'brun.generators',
        'brun.combinators',
        'brun.utils',
    ],
    version=lib_version,
    license='MIT',
    description='Utility used to run parameterized shell commands',
    author='Andrea F. Daniele',
    author_email='afdaniele@ttic.edu',
    url='https://github.com/afdaniele/',
    download_url='https://github.com/afdaniele/run-batch/tarball/{}'.format(lib_version),
    zip_safe=False,
    include_package_data=True,
    keywords=['batch', 'parameterized', 'commands', 'shell'],
    install_requires=['networkx'],
    scripts=['brun/brun'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



import os
import json

aliases = ['j']


def generate(args):
    if len(args) != 1:
        msg = 'The field type "json" takes exactly one argument. i.e., a path to a json file'
        raise ValueError(msg)
    # ---
    # get file path
    filepath = os.path.join(os.getcwd(), args[0])
    # make sure the file exists
    if not os.path.exists(filepath):
        msg = 'The file "{}" does not exist.'.format(filepath)
        raise FileNotFoundError(msg)
    # load file from disk
    try:
        with open(filepath, 'r') as fin:
            data = json.load(fin)
    except json.decoder.JSONDecodeError as e:
        msg = 'The file "{}" must contain a valid JSON. Error: {}'.format(filepath, str(e))
        raise json.decoder.JSONDecodeError(msg, e.doc, e.pos)
    # simple case: empty json
    if len(data) == 0:
        return {}
    _, lst = next(iter(data.items()))
    # check if the file contains a signle key
    if not (len(data) == 1 and isinstance(lst, list)):
        msg = 'The file "{}" must contain a single key with a list of objects as value'.format(
            filepath)
        raise JSONInputError(msg)
    # simple case: empty list
    if len(lst) == 0:
        return {}
    # check if the list contains dictionaries only
    if sum([int(isinstance(e, dict)) for e in lst]) != len(lst):
        msg = 'The file "{}" must contain a single key with a list of objects as value'.format(
            filepath)
        raise JSONInputError(msg)
    # get the prototype of the objects from the first one
    _proto = lambda d: [(k, type(v)) for k, v in d.items()]
    proto = _proto(lst[0])
    # allowed types are [int, float, string]
    for _, t in proto:
        if t not in [int, float, str]:
            msg = 'All the objects in the file "{}" must contain only values of types [int, float, str]'.format(
                filepath)
            raise JSONInputError(msg)
    # check if all the dictionaries contain the same keys => types
    for d in lst:
        if _proto(d) != proto:
            msg = 'All the objects in the file "{}" must share the same prototype'.format(filepath)
            raise JSONInputError(msg)
    # get values
    keys = [k for k, _ in proto]
    values = {k: [] for k in keys}
    for d in lst:
        for k, v in d.items():
            values[k].append(v)
    # ---
    return values


class JSONInputError(RuntimeError):
    def __init__(self, msg):
        super(RuntimeError, self).__init__(msg)
